;;; -*- lisp -*-
;;; $Id: webview.lisp 432 2003-08-22 22:14:05Z miles $

(in-package :macho)

(defmacro who (stream &body body)
  `(with-html-output (,stream nil :prologue nil :indent *indent*)
    ,@body))

(defun page-head (stream title stylesheets)
  (who stream
       (:head
        (:title (str title))
        (loop for s in stylesheets
              do (htm (:link :rel "stylesheet" :type "text/css" :href s))))))

(defun write-archive-index (archive path)
  (msg "writing archive index to ~A~%" path)
  (ensure-directories-exist path)
  (with-slots (name messages year-table) archive
    (let ((year (this-year))
          (month (this-month))
          (years (loop for k being each hash-key of year-table collect k)))
      (setf years (sort years #'>))
      (with-open-file (index path :direction :output)
        (who index
             (:html
              (page-head index (s+ name " mail archive") '("../../styles.css"))
              (:body
               (:div :class "page-header"
                     (:a :class "logo" :href "../"
                         (:span :class "name" (str (archive-name archive)))
                         (:span :class "word" "archives")))
               (:table :class "archive-index"
                       (:th)
                       (loop for i from 0 to 11
                             do (htm (:th (fmt "~3,3/net.telent.date:monthname/" (1+ i)))))
                       (loop for year in years
                             for msgs = (gethash year year-table)
                             do (htm (:tr
                                      (:td :class "year" (str year))
                                      (loop
                                       for i from 0 to 11
                                       for m = (svref msgs i)
                                       for count = (length m)
                                       do (if (> count 0)
                                              (htm (:td (:a :href (s+ year "-" (1+ i) "/index.html") (str count))))
                                              (htm (:td "&nbsp;")))))))))))))))

(defun month-index-sort-controls (s)
  (who s
       (:div :class "nav"
             (:span :class "control" (:a :href "../index.html" "up"))
             (loop for index in '("thread" "author" "date" "subject")
                   do (htm (:span :class "control"
                                  (:a :class "first" :href (s+ index "-descending.html") (str (s+ "sort by " index)))
                                  (:a :href (s+ index "-ascending.html") "&lt;")))))))

(defun message-path (message)
  (with-slots (time serial) message
    (s+ "../" (time-year time) "-" (time-month time) "/" serial ".html")))
  
(defun month-index-entry (msg num stream)
  (with-slots (serial subject author time) msg
    (who stream
         (:span
          (:a :class "shortcut" :href (message-path msg) (str num))
          (:a :href (message-path msg) (str subject))
          (:span :class "author" (str author))
          (:span :class "date" (fmt "~A/~A ~D:~2,'0D"
                                    (time-month time)
                                    (time-date time)
                                    (time-hour time)
                                    (time-minute time)))))))

(defun write-month-index (archive year month path)
  (msg "writing month index for ~A / ~A to ~A~%" year (1+ month) path)
  (ensure-directories-exist path)
  (with-open-file (index path :direction :output)
    (with-slots (name messages year-table) archive
      (who index
           (:html
            (page-head index
                       (format nil "~A mail archive - ~/net.telent.date:monthname/ ~A" name (1+ month) year)
                       '("../../../styles.css"))
            (:body
             (:div :class "page-header"
                   (:a :class "logo" :href "../"
                       (:span :class "name" (str (archive-name archive)))
                       (:span :class "word" "archives"))
                   (:span :class "logo-date" (str (s+ year "-" (1+ month))))
                   (month-index-sort-controls index))
             (:ul :class "message-list"
                  (loop for msg in (svref (gethash year year-table) month)
                        for i upfrom 1
                        do (htm (:li (month-index-entry msg i index)))))))))))

(defun write-month-thread-index (archive year month path)
  (msg "writing thread index for ~A / ~A to ~A~%" year (1+ month) path)
  (ensure-directories-exist path)
  (let ((current 0))
    (labels ((render-msg (msg s)
               (who s
                    (:li
                     (str (month-index-entry msg current s))
                     (if (message-children msg)
                         (htm (:ul
                               (loop for child in (message-children msg)
                                     do (incf current)
                                     do (render-msg child s)))))))))
      (with-open-file (index path :direction :output)
        (with-slots (name messages year-table) archive
          (who index
               (:html
                (page-head index
                           (format nil "~A mail archive - ~/net.telent.date:monthname/ ~A" name (1+ month) year)
                           '("../../../styles.css"))
                (:body
                 (:div :class "page-header"
                       (:a :class "logo" :href "../"
                           (:span :class "name" (str (archive-name archive)))
                           (:span :class "word" "archives"))
                       (:span :class "logo-date" (str (s+ year "-" (1+ month))))
                       (month-index-sort-controls index))
                 (:ul :class "message-list"
                      (loop for msg in (svref (gethash year year-table) month)
                            if (not (and (message-parent msg)
                                         (= (time-month (message-time (message-parent msg)))
                                            (1+ month))))
                            do (progn
                                 (incf current)
                                 (render-msg msg index))))))))))))

(defun write-message (path msg &optional prev next)
  (ensure-directories-exist path)
  (with-open-file (stream path :direction :output)
    (with-slots (archive author children serial subject time id in-reply) msg
      (msg "writing msg ~S ~S ~S ~S~%" serial author subject id)
      (who stream
           (:html
            (page-head stream
                       (format nil "~A mail archive - : ~A (~A)" (archive-name archive) subject author)
                       '("../../../styles.css"))
            (:div :class "page-header"
                  (:a :class "logo" :href "../"
                      (:span :class "name" (str (archive-name archive)))
                      (:span :class "word" "archives"))
                  (:span :class "logo-date" (str (s+ (time-year time) "-" (time-month time) "-" (time-date time))))
                  (:div :class "nav"
                        (:span :class "control"
                               (:a :href "index.html" :title "up to index" "up"))
                        (:span :class "control"
                               (if prev
                                   (htm (:a :href (message-path prev) :title "previous message" "prev <"))
                                   (htm (:span "prev <"))))
                        (:span :class "control"
                               (if next
                                   (htm (:a :href (message-path next) :title "next message" "next >"))
                                   (htm (:span "next >"))))
                        (:span :class "control"
                               (if children
                                   (let ((c (first children)))
                                     (htm (:a :href (s+ (message-serial (first children)) ".html") "thread-next >>")))
                                   (htm :span "thread-next >>")))))
            (:div :class "message-headers"
                  (:p (str (s+ "From: " author)))
                  (:p (str (s+ "Subject: " subject)))
                  (:p (str (format nil "Date: ~A-~A-~A ~D:~2,'0D"
                                   (time-year time)
                                   (time-month time)
                                   (time-date time)
                                   (time-hour time)
                                   (time-minute time)))))
            (if (message-body msg)
                (htm (:pre :class "message-body"
                           (str (obscure-email-addresses (message-body msg)))))))))))

(defun write-archive (archive path)
  (msg "building web pages for ~S~%" (archive-name archive))
  (with-slots (name messages year-table dirty-table) archive
    (write-archive-index archive (join-path path "index.html"))
    (loop for year being each hash-key of year-table
          using (hash-value msgs)
          do (if (gethash year dirty-table)
                 (progn
                   (dotimes (month 12)
                     (let ((month-dir (join-path path (s+ year "-" (1+ month))))
                           (month-dirty (svref (gethash year dirty-table) month)))
                       (if month-dirty
                           (progn
                             (loop for name in (list "author" "date" "subject")
                                   for pred in (list #'string< #'time< #'string<)
                                   for key in (list #'message-author #'message-time #'message-subject)
                                   do (loop for suffix in (list "ascending" "descending")
                                            for pred2 in (list pred #'(lambda (a b) (not (funcall pred a b))))
                                            do (progn
                                                 (setf (svref msgs month)
                                                       (stable-sort (svref msgs month) pred2 :key key))
                                                 (write-month-index archive
                                                                    year
                                                                    month
                                                                    (join-path month-dir (s+ name "-" suffix ".html"))))))
                             (setf (svref msgs month)
                                   (stable-sort (svref msgs month) #'time< :key #'message-time))
                             (write-month-thread-index archive
                                                       year
                                                       month
                                                       (join-path month-dir "thread-ascending.html"))
                             (setf (svref msgs month)
                                   (stable-sort (svref msgs month) #'time> :key #'message-time))
                             (write-month-thread-index archive
                                                       year
                                                       month
                                                       (join-path month-dir "thread-descending.html"))
                             (let ((index-path (join-path month-dir "index.html")))
                               (if (probe-file index-path)
                                   (sb-posix:unlink index-path))
                               (sb-posix:symlink "thread-descending.html" index-path))
                             (if (and (= (1+ month) (this-month)) (= year (this-year)))
                                 (let ((latest-path (join-path path "latest")))
                                   (if (probe-file latest-path)
                                       (sb-posix:unlink (join-path path "latest")))
                                   (sb-posix:symlink (s+ year "-" (1+ month)) latest-path)))
                             (loop with msgs = (svref msgs month)
                                   with last = (length messages)
                                   for msg in msgs
                                   for serial = (message-serial msg)
                                   do (if (message-dirty msg)
                                          (write-message (join-path month-dir (s+ serial ".html"))
                                                         msg
                                                         (if (= serial 1) nil (aref messages (- serial 2)))
                                                         (if (= serial last) nil (aref messages serial))))))))))))))

